import json
import re
import sys
from base64 import b64decode

sys.path.insert(0, '/motadata/motadata/motadata python engine')

from com.motadata.commons.sshclient import TraceOrgSSHClient
from com.motadata.util.loggerutil import TraceOrgLogger

_logger = TraceOrgLogger("Linux Metric")

def collect(context):

    metrics = {}

    if context and context.get('plugin-context'):

        plugin_context = context.get("plugin-context")

        ssh_client = None

        host = None

        timeout = 60

        if context.get("metric-collect-timeout"):

            timeout = convert_to_int(context.get("metric-collect-timeout"))

        plugin_context['timeout'] = timeout

        try:

            if plugin_context.get('username') and plugin_context.get('host'):

                _logger.debug("metric collection request:" + plugin_context.get("host"))

                host = convert_to_str(plugin_context.get('host'))

                ssh_client = get_ssh_client(plugin_context, host)

                if ssh_client:

                    if ssh_client.init():

                        user_name = convert_to_str(plugin_context.get('username'))

                        set_cpu_details(ssh_client, metrics)

                        metrics['Disk Volumes'] = set_logical_disk_details(metrics, ssh_client)

                        set_memory_details(metrics, ssh_client)

                        set_uptime(ssh_client, metrics)

                        set_disk_io_details(ssh_client, metrics)

                        set_network_card_details(ssh_client, metrics)

                        set_running_process_details(ssh_client, metrics)

                        set_total_thread_counf(ssh_client, metrics)

                        set_vendor_model_details(ssh_client, metrics)

                        set_miscellaneous_details(ssh_client, metrics)

                        set_password_expiry_days(ssh_client,metrics,user_name)

                        if plugin_context.get('processes') and len(convert_to_str(plugin_context.get('processes'))) > 0:
                            set_process_details(ssh_client, convert_to_str(plugin_context.get('processes')), metrics)

                        if plugin_context.get('process-ids') and len(convert_to_str(plugin_context.get('process-ids'))) > 0:

                            processes = []

                            for process_id in convert_to_str(plugin_context.get("process-ids")).split(","):
                                set_process_details_by_process_id(ssh_client, convert_to_int(process_id), processes,
                                                                  metrics)

                            if len(processes) > 0:
                                metrics['Processes'] = processes

                        if plugin_context.get('services') and len(convert_to_str(plugin_context.get('services'))) > 0:
                            get_service_details(ssh_client, convert_to_str(plugin_context.get('services')), metrics)

                        _logger.debug("metric details of {} : {}".format(host, json.dumps(metrics)))

                    else:
                        metrics['error-code'] = 'Invalid Credentials {}'.format(host)
                else:
                    metrics['error-code'] = 'No Valid Credentials found {}'.format(host)
            else:
                metrics['error-code'] = 'No valid credentials found {}'.format(host)

        except Exception as exception:

            _logger.warn("failed to collect linux metric " + host)

            metrics['error-code'] = exception.__class__.__name__

        finally:
            if ssh_client:
                ssh_client.destroy()

    return metrics


def get_ssh_client(plugin_context, host):
    ssh_client = None

    try:

        user_name = convert_to_str(plugin_context.get('username'))

        password = convert_to_str(plugin_context.get('password'))

        port = convert_to_int(plugin_context.get('port'))

        timeout = convert_to_int(plugin_context['timeout'])

        if plugin_context.get('private-key-name') and len(convert_to_str(plugin_context.get('private-key-name'))) > 0:

            private_key = convert_to_str(plugin_context.get('private-key-name'))

            passphrase = None

            if plugin_context.get('pass-phrase') and len(convert_to_str(plugin_context.get('pass-phrase'))) > 0:

                passphrase = convert_to_str(plugin_context.get('pass-phrase'))

            ssh_client = TraceOrgSSHClient(host=host,
                                           port=port,
                                           username=user_name,
                                           password=password,
                                           timeout=timeout,
                                           key_file=private_key,
                                           passphrase=passphrase)

        else:

            ssh_client = TraceOrgSSHClient(host=host,
                                           port=port,
                                           username=user_name,
                                           password=password,
                                           timeout=timeout)


    except Exception:
        _logger.error()

    return ssh_client


def set_cpu_details(ssh_client, metrics):
    try:

        if ssh_client:

            output = ssh_client.execute_command("mpstat -P ALL | awk '{print $11}'")

            if output and len(str(output)) > 0 and 'idle' in output:

                output = ssh_client.execute_command(
                    "mpstat -P ALL 1 1 | grep -vE -i 'Average'| awk '{print $11\"" + " " + "\"$4\"" + " " + "\"$6\"" + " " + "\"$7\"" + " " + "\"$8\"" + " " + "\"$3 }'")

                set_cpu_counter_details(metrics, output, 'mpstat', ssh_client)

            else:
                output = ssh_client.execute_command("mpstat -P ALL | awk '{print $12}'")

                if output and len(str(output)) > 0 and 'idle' in output:

                    output = ssh_client.execute_command(
                        "mpstat -P ALL 1 1 | grep -vE -i 'Average'| awk '{print $12\"" + " " + "\"$4\"" + " " + "\"$6\"" + " " + "\"$7\"" + " " + "\"$8\"" + " " + "\"$3}'")

                    set_cpu_counter_details(metrics, output, 'mpstat', ssh_client)

                else:
                    output = ssh_client.execute_command("mpstat -P ALL | awk '{print $13}'")

                    if output and len(str(output)) > 0 and 'idle' in output:

                        output = ssh_client.execute_command(
                            "mpstat -P ALL 1 1 | grep -vE -i 'Average'| awk '{print $13\"" + " " + "\"$4\"" + " " + "\"$6\"" + " " + "\"$7\"" + " " + "\"$8\"" + " " + "\"$3}'")

                        set_cpu_counter_details(metrics, output, 'mpstat', ssh_client)

                    else:

                        output = ssh_client.execute_command(
                            "vmstat 1 3 | grep -v procs | grep -v r | awk {'print $13\"" + " " + "\"$14\"" + " " + "\"$15'}")

                        set_cpu_counter_details(metrics, output, 'non-mpstat', ssh_client)


    except Exception:

        _logger.error()


def set_cpu_counter_details(metrics, output, cpu_type, ssh_client):
    cpu_cores = []

    processors = 0

    total_cpu_idle = 0

    total_cpu_user = 0

    total_cpu_system = 0

    total_cpu_io = 0

    total_cpu_interrupt = 0

    try:

        if cpu_type == 'mpstat':

            if output and len(str(output)) > 0 and str(output) != "\n":

                pattern = re.compile(
                    "(\\d+\\.\\d+)\\s(\\d+\\.\\d+)\\s(\\d+\\.\\d+)\\s(\\d+\\.\\d+)\\s(\\d+\\.\\d+)\\s(\\d+)")

                for matcher in re.finditer(pattern, output):
                    row = {}

                    processors = processors + 1

                    row["CPU Core Idle (%)"] = convert_to_int(matcher.group(1))

                    row["CPU Core User (%)"] = convert_to_int(matcher.group(2))

                    row["CPU Core System (%)"] = convert_to_int(matcher.group(3))

                    row["CPU Core I/O (%)"] = convert_to_int(matcher.group(4))

                    row["CPU Core Interrupt (%)"] = convert_to_int(matcher.group(5))

                    row["CPU Core Utilization (%)"] = 100 - convert_to_int((row["CPU Core Idle (%)"]))

                    row["CPU Core"] = convert_to_int(matcher.group(6))

                    total_cpu_idle += convert_to_int(matcher.group(1))

                    total_cpu_user += convert_to_int(matcher.group(2))

                    total_cpu_system += convert_to_int(matcher.group(3))

                    total_cpu_io += convert_to_int(matcher.group(4))

                    total_cpu_interrupt += convert_to_int(matcher.group(5))

                    cpu_cores.append(row)

                if processors > 0:

                    if len(cpu_cores) > 0:
                        metrics['CPU Logical Cores'] = cpu_cores

                    cpu_average = ssh_client.execute_command(
                        "mpstat  1 1 | grep 'Average'| awk '{print $3\"" + " " + "\"$5\"" + " " + "\"$6\"" + " " + "\"$7\"" + " " + "\"$12}'")

                    pattern = re.compile(
                        "(\\d+\\.\\d+)\\s(\\d+\\.\\d+)\\s(\\d+\\.\\d+)\\s(\\d+\\.\\d+)\\s(\\d+\\.\\d+)")

                    matcher = None

                    if cpu_average:

                        matcher = pattern.match(cpu_average.replace("\n", ""))

                    if matcher:

                        metrics["User CPU (%)"] = convert_to_int(matcher.group(1))

                        metrics["System CPU (%)"] = convert_to_int(matcher.group(2))

                        metrics["I/O CPU (%)"] = convert_to_int(matcher.group(3))

                        metrics["Interrupt CPU (%)"] = convert_to_int(matcher.group(4))

                        metrics["Idle CPU (%)"] = convert_to_int(matcher.group(5))

                        metrics["CPU (%)"] = 100 - convert_to_int(metrics.get("Idle CPU (%)"))

                    else:

                        metrics["Idle CPU (%)"] = convert_to_int(total_cpu_idle / processors)

                        metrics["User CPU (%)"] = convert_to_int(total_cpu_user / processors)

                        metrics["System CPU (%)"] = convert_to_int(total_cpu_system / processors)

                        metrics["I/O CPU (%)"] = convert_to_int(total_cpu_io / processors)

                        metrics["Interrupt CPU (%)"] = convert_to_int(total_cpu_interrupt / processors)

                        metrics["CPU (%)"] = 100 - convert_to_int(metrics.get("Idle CPU (%)"))

        else:

            if output and len(str(output)) > 0 and str(output) != "\n":

                cpu_output = output.split("\n")

                if len(cpu_output) >= 3:
                    output = cpu_output[2].strip()

                    cpu_output = output.split(" ")

                    metrics["User CPU (%)"] = convert_to_int(cpu_output[0] if cpu_output[0] else 0)

                    metrics["System CPU (%)"] = convert_to_int(cpu_output[1] if cpu_output[0] else 0)

                    metrics["Idle CPU (%)"] = convert_to_int(cpu_output[2] if cpu_output[0] else 0)

                    metrics["CPU (%)"] = 100 - convert_to_int(metrics.get("Idle CPU (%)"))


    except Exception:

        _logger.error()


def set_logical_disk_details(metric, ssh_client):
    disk_usage_details = []

    overall_usage = 0

    overall_free = 0

    overall_size = 0

    overall_percentage = 0

    try:
        if ssh_client:

            output = ssh_client.execute_command(
                "df -l -k | grep -vE '^Filesystem|tmpfs|cdrom|none' | awk '{ print $1\"" + " " + "\"$2\"" + " " + "\"$3\"" + " " + "\"$4\"" + " " + "\"$5\"" + " " + "\"$6}'")

            if output and len(output) > 0 and output != '\n':

                pattern = re.compile("(.*)[\\s\\n\\r]+(\\d+)[\\s]+(\\d+)[\\s]+(\\d+)[\\s]+(\\d+)%[\\s]+(.*)")

                for matcher in re.finditer(pattern, output):
                    row = {}

                    volume_capacity = convert_to_int(convert_to_int(matcher.group(2)) / 1024)

                    if volume_capacity > 0:
                        row["Disk Volume"] = convert_to_str(matcher.group(1))

                        row["Disk Volume Capacity (MB)"] = convert_to_int(convert_to_int(matcher.group(2)) / 1024)

                        row["Disk Volume Used (MB)"] = convert_to_int(convert_to_int(matcher.group(3)) / 1024)

                        row["Disk Volume Free (MB)"] = convert_to_int(convert_to_int(matcher.group(4)) / 1024)

                        row["Disk Volume Utilization (%)"] = convert_to_int(convert_to_int(matcher.group(5)))

                        row["Disk Volume Mount Path"] = convert_to_str(matcher.group(6))

                        overall_size += convert_to_int(row.get("Disk Volume Capacity (MB)"))

                        overall_usage += convert_to_int(row.get("Disk Volume Used (MB)"))

                        overall_free += convert_to_int(row.get("Disk Volume Free (MB)"))

                        overall_percentage += convert_to_int(row.get("Disk Volume Utilization (%)"))

                        disk_usage_details.append(row)

                if overall_size > 0 and overall_usage > 0 and len(disk_usage_details) > 0:
                    overall_percentage = convert_to_int(overall_usage * 100 / overall_size)

                    metric["Disk Capacity (MB)"] = overall_size

                    metric["Used Disk (MB)"] = overall_usage

                    metric["Free Disk (MB)"] = overall_free

                    metric["Disk (%)"] = overall_percentage

    except Exception:
        _logger.error()

    return disk_usage_details


def set_memory_details(metrics, ssh_client):
    try:

        if ssh_client:

            output = ssh_client.execute_command(
                "free -m |grep -i mem |awk '{print $2\" " + " " + "\" $3\" " + " " + "\" $6\" " + " " + "\" $7}'")

            if output and len(output) > 0 and output != '\n':

                memory_outputs = output.replace("\r", "").split()

                total_mem = convert_to_int(memory_outputs[0] if memory_outputs[0] else 0)

                used_mem = convert_to_int(memory_outputs[1] if memory_outputs[1] else 0)

                buff_mem = convert_to_int(memory_outputs[2] if memory_outputs[2] else 0)

                buff_mem += convert_to_int(memory_outputs[3] if memory_outputs[3] else 0)

                used_mem = used_mem - buff_mem

                if used_mem > 0:

                    per_usage = 0

                    if total_mem > 0:
                        per_usage = convert_to_int((used_mem * 100) / total_mem)

                    free_mem = total_mem - used_mem

                    metrics["Installed Memory (MB)"] = total_mem

                    metrics["Used Memory (MB)"] = used_mem

                    metrics["Free Memory (MB)"] = free_mem

                    metrics["Memory (%)"] = per_usage

                    metrics["Overall Used Memory (MB)"] = used_mem + buff_mem

                    metrics["Overall Free Memory (MB)"] = total_mem - (used_mem + buff_mem)

                    if total_mem > 0:
                        metrics["Overall Memory (%)"] = convert_to_int(((used_mem + buff_mem) * 100) / total_mem)

                else:

                    output = ssh_client.execute_command(
                        "free -m |grep -i mem |awk '{print $2\" " + " " + "\" $3\" " + " " + "\" $4\" " + " " + "\" $6\" " + " " + "\" $7}'")

                    memory_outputs = output.replace("\r", "").split()

                    total_mem = convert_to_int(memory_outputs[0] if memory_outputs[0] else 0)

                    used_mem = convert_to_int(memory_outputs[1] if memory_outputs[1] else 0)

                    free_mem = convert_to_int(memory_outputs[2] if memory_outputs[2] else 0)

                    buff_mem = convert_to_int(memory_outputs[3] if memory_outputs[3] else 0)

                    avail_mem = convert_to_int(memory_outputs[4] if memory_outputs[4] else 0)

                    per_usage = 0

                    if total_mem > 0:
                        per_usage = convert_to_int(((used_mem * 100) / total_mem))

                    metrics["Installed Memory (MB)"] = total_mem

                    metrics["Used Memory (MB)"] = used_mem

                    metrics["Free Memory (MB)"] = avail_mem

                    metrics["Memory (%)"] = per_usage

                    metrics["Overall Used Memory (MB)"] = used_mem

                    metrics["Overall Free Memory (MB)"] = free_mem

                    if total_mem > 0:
                        metrics['Overall Memory (%)'] = convert_to_int(((used_mem + buff_mem) * 100) / total_mem)

            output = ssh_client.execute_command("free -w |grep -i mem |awk '{print $6\" " + " " + "\" $7}'")

            if output and len(output) > 0 and output != '\n':

                memory_outputs = output.replace("\r", "").split()

                total_mem = convert_to_int(memory_outputs[0] if memory_outputs[0] else 0)

                used_mem = convert_to_int(memory_outputs[1] if memory_outputs[1] else 0)

                metrics["Buffered Memory (MB)"] = convert_to_int(total_mem / 1024)

                metrics["Cached Memory (MB)"] = convert_to_int(used_mem / 1024)

            else:

                output = ssh_client.execute_command("free -m |grep -i mem |awk '{print $6\" " + " " + "\" $7}'")

                if output and len(output) > 0 and output != '\n':
                    memory_outputs = output.replace("\r", "").split()

                    total_mem = convert_to_int(memory_outputs[0] if memory_outputs[0] else 0)

                    used_mem = convert_to_int(memory_outputs[1] if memory_outputs[1] else 0)

                    metrics["Buffered Memory (MB)"] = convert_to_int(total_mem / 1024)

                    metrics["Cached Memory (MB)"] = convert_to_int(used_mem / 1024)

            output = ssh_client.execute_command("free -m | grep -i swap |awk '{print $2\" " + " " + "\" $3}'")

            if output and len(output) > 0 and output != '\n':

                memory_outputs = output.replace("\r", "").split()

                total_mem = convert_to_int(memory_outputs[0] if memory_outputs[0] else 0)

                used_mem = convert_to_int(memory_outputs[1] if memory_outputs[1] else 0)

                if total_mem == 0:
                    per_usage = 0
                else:
                    per_usage = convert_to_int((used_mem * 100) / total_mem)

                metrics["Swap Memory (MB)"] = total_mem

                metrics["Used Swap Memory (MB)"] = used_mem

                metrics["Free Swap Memory (MB)"] = (total_mem - used_mem) if total_mem > 0 else 0

                metrics["Swap Memory (%)"] = per_usage

    except Exception:
        _logger.error()


def set_uptime(ssh_client, metrics):
    try:
        output = ssh_client.execute_command("cat /proc/uptime")

        if output and len(output) > 0 and output != "\n":
            metrics["System Uptime (Seconds)"] = convert_to_int(output.split(" ")[0].strip())

    except Exception:
        _logger.error()


def set_disk_io_details(ssh_client, metrics):
    disk_io_details = []

    try:
        output = ssh_client.execute_command("iostat -d -xk | grep -vE 'Linux|Device' | awk '{print $14}'")

        if output and len(output.replace("\n", "").replace("\r", "").strip()) > 0:

            output = ssh_client.execute_command(
                "iostat -d -xk | grep -vE 'Linux|Device' |awk '{print $1\"" + " " + "\"$4\"" + " " + "\"$5\"" + " " + "\"$6\"" + " " + "\"$7\"" + " " + "\"$9\"" + " " + "\"$14}'")

        else:

            output = ssh_client.execute_command(
                "iostat -d -xk | grep -vE 'Linux|Device' |awk '{print $1\"" + " " + "\"$4\"" + " " + "\"$5\"" + " " + "\"$6\"" + " " + "\"$7\"" + " " + "\"$9\"" + " " + "\"$12}'")

        if output and len(output.replace("\n", "").replace("\r", "").strip()) > 0:

            pattern = re.compile(
                "(.+)\\s(\\d+)(?:\\.\\d+)\\s(\\d+)(?:\\.\\d+)\\s(\\d+)(?:\\.\\d+)\\s(\\d+)(?:\\.\\d+)\\s(\\d+)(?:\\.\\d+)\\s(\\d+)(?:\\.\\d+)")

            for matcher in re.finditer(pattern, output):
                row = {}

                row["Disk"] = convert_to_str(matcher.group(1))

                row["Disk Reads/sec"] = convert_to_int(matcher.group(2))

                row["Disk Writes/sec"] = convert_to_int(matcher.group(3))

                row["Disk I/O Ops/sec"] = convert_to_int(matcher.group(2)) + convert_to_int(matcher.group(3))

                row["Disk Byte Reads/sec"] = convert_to_int(matcher.group(4)) * 1024

                row["Disk Byte Writes/sec"] = convert_to_int(matcher.group(5)) * 1024

                row["Disk Bytes/sec"] = convert_to_int(row["Disk Byte Writes/sec"]) + convert_to_int(
                    row["Disk Byte Reads/sec"])

                row["Disk Queue Length"] = convert_to_int(matcher.group(6))

                row["Disk I/O Time (%)"] = convert_to_int(matcher.group(7))

                disk_io_details.append(row)

            if len(disk_io_details) > 0:
                metrics['Disks'] = disk_io_details

    except Exception:
        _logger.error()


def set_network_card_details(ssh_client, metrics):
    try:
        output = ssh_client.execute_command("/sbin/ifconfig ")

        metrics['netword-card'] = output

    except Exception:

        _logger.error()


def set_running_process_details(ssh_client, metrics):
    try:
        output = ssh_client.execute_command("ps aux | wc -l")

        if output and len(output) > 0 and output != "\n":
            metrics["Running Processes"] = convert_to_int(output.split(" ")[0].strip())

    except Exception:
        _logger.error()


def set_total_thread_counf(ssh_client, metrics):
    try:
        output = ssh_client.execute_command(
            "ps -eo nlwp | tail -n +2 | awk '{ num_threads += $1 } END { print num_threads }'")

        if output and len(output) > 0 and output != "\n":
            metrics["Threads"] = convert_to_int(output.split(" ")[0].strip())

    except Exception:
        _logger.error()


def set_vendor_model_details(ssh_client, metrics):
    try:
        output = ssh_client.execute_command("cat /proc/cpuinfo | awk '/vendor_id/ || /model name/'")

        if output and len(output) > 0 and output != "\n":

            counter_values = output.split("\t")

            for i in range(len(counter_values)):

                if "vendor_id" in counter_values[i]:
                    metrics["Vendor"] = convert_to_str(counter_values[i + 1].replace(":", "").split("\n")[0])

                if "model name" in counter_values[i]:
                    metrics["Model"] = convert_to_str(counter_values[i + 1].replace(":", "").split("\n")[0])

                    break


    except Exception:
        _logger.error()


def set_miscellaneous_details(ssh_client, metrics):
    try:

        output = ssh_client.execute_command("uname -o")

        if output and len(output) > 0 and output != "\n":
            metrics["OS Name"] = convert_to_str(output)

        output = ssh_client.execute_command("uname -p")

        if output and len(output) > 0 and output != "\n":
            metrics["CPU Type"] = convert_to_str(output)

        output = ssh_client.execute_command("uname -n")

        if output and len(output) > 0 and output != "\n":
            metrics["System Name"] = convert_to_str(output)

        output = ssh_client.execute_command("uname -r")

        if output and len(output) > 0 and output != "\n":
            metrics["OS Version"] = convert_to_str(output)

        output = ssh_client.execute_command("vmstat | awk '{print $1\" \"$2\" \"$11\" \"$12}'")

        if output and len(output) > 0 and output != "\n":

            pattern = re.compile("(?:.*)[\\s]+(?:.*)[\\s]+(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)")

            for matcher in re.finditer(pattern, output):
                metrics["Processor Queue Length"] = convert_to_int(matcher.group(1))

                metrics["Blocked Processes"] = convert_to_int(matcher.group(2))

                metrics["Interrupts/sec"] = convert_to_int(matcher.group(3))

                metrics["Context Switches/sec"] = convert_to_int(matcher.group(4))

        output = ssh_client.execute_command("cat /proc/sys/fs/file-nr | awk '{print $1}'")

        if output and len(output) > 0 and output != "\n":
            metrics["Open File Descriptors"] = convert_to_int(output)

        output = ssh_client.execute_command("cat /proc/cpuinfo | grep processor | wc -l")

        if output and len(output) > 0 and output != "\n":
            metrics["CPU Cores"] = convert_to_int(output)


    except Exception:

        _logger.error()


def set_process_details_by_process_id(ssh_client, process_id, processes, metrics):
    process_memory = 0

    memory_output = None

    try:

        pattern = re.compile("(\\w.*)\\s+(\\d+)\\s+(\\d+.\\d+)\\s+(\\d+.\\d+)\\s+(.*)")

        io_disk_pattern = re.compile("(\\d+)\\s+(\\d+)(?:.\\d+)\\s+(\\d+)(?:.\\d+)")

        output = ssh_client.execute_command(
            "top -b -n 1 -p " + process_id + " | awk '{print $2\" \"$1\" \"$9\" \"$10\" \"$12}'")

        if output and len(output) > 0 and output != "\n":

            for matcher in re.finditer(pattern, output):

                row = {}

                process_name = convert_to_str(matcher.group(5))

                row['Process'] = process_name

                row["Process User"] = convert_to_str(matcher.group(1))

                row["Process Id"] = convert_to_str(process_id)

                output = ssh_client.execute_command("ps hH p " + convert_to_str(matcher.group(2)) + " | wc -l")

                if output and len(output) > 0 and output != "\n":
                    row["Process Threads"] = convert_to_int(output.strip())

                process_cpu = convert_to_float(matcher.group(3).strip())

                cpu_cores = convert_to_int(metrics.get("CPU Cores"))

                if cpu_cores > 0:
                    row["Process CPU (%)"] = convert_to_int(process_cpu / cpu_cores)

                row["Process CPU Time (%)"] = convert_to_int(process_cpu)

                if memory_output and memory_output != "":
                    process_memory = convert_to_int(memory_output.strip()) * 1024

                if process_memory != -1:
                    process_memory_kb = (convert_to_float(matcher.group(4)) * process_memory) / 100

                    row["Process Memory (KB)"] = convert_to_int(process_memory_kb)

                output = ssh_client.execute_command("pidstat -p " + process_id + " -d | awk '{print $4\" \"$5\" \"$6}'")

                if output and len(output) > 0 and output != "\n":

                    for matcher_io_disk in re.finditer(io_disk_pattern, output):

                        if convert_to_str(row.get("Process Id")) == convert_to_str(matcher_io_disk.group(1)):
                            process_disk_reads = convert_to_int(matcher_io_disk.group(2))

                            process_disk_writes = convert_to_int(matcher_io_disk.group(3))

                            row["Process I/O (Bytes/sec)"] = (process_disk_reads + process_disk_writes) * 1024

                if len(row) > 0:
                    processes.append(row)

    except Exception:
        _logger.error()


def set_process_details(ssh_client, process, metrics):
    process_list = []

    try:

        if process == '*':

            process_details = ssh_client.execute_command("top -b -n 1 | awk '{print $12}'")

            if process_details and len(convert_to_str(process_details)) > 0:

                process_names = process_details.split("\n")

                command_found = False

                for process_name in process_names:

                    if str(process_name).lower() == "command":
                        command_found = True

                    if command_found and str(process_name).lower() != "command":
                        process_list.append(str(process_name))

            if len(process_list) > 0:
                process_list = "\|".join(process_list)

        else:
            process_list = process.replace("*", "").replace(",", "\|")

        if process_list and len(convert_to_str(process_list)) > 0:

            process_details = ssh_client.execute_command(
                "top -b -n 1 |grep '" + process_list + "' | awk '{print $2\" \"$1\" \"$9\" \"$10\" \"$12}'")

            if process_details and len(convert_to_str(process_details)) > 0:
                metrics['process-details'] = process_details

            process_memory_details = ssh_client.execute_command("free -m |grep -i mem |awk '{print $2}'")

            if process_memory_details and len(convert_to_str(process_memory_details)) > 0:
                metrics['process-memory-details'] = process_memory_details

            process_state_details = ssh_client.execute_command("pidstat -d | awk '{print $4\" \"$5\" \"$6}'")

            if process_state_details and len(convert_to_str(process_state_details)) > 0:
                metrics['process-state-details'] = process_state_details

            process_thread_details = ssh_client.execute_command("ps -e -T | awk '{print $1}'")

            if process_thread_details and len(convert_to_str(process_thread_details)) > 0:
                metrics['process-thread-details'] = process_thread_details

    except Exception:

        _logger.error()


def get_service_details(ssh_client, services, metrics):
    services_metrics = []

    try:

        if services and len(convert_to_str(services)) > 0:

            if convert_to_str(services) != "*":

                service_list = get_services(ssh_client, services_metrics)

                for service_name in convert_to_str(services).split(","):

                    if "*" in service_name:

                        final_service_details = list(services_metrics)

                        for service_list_name in service_list:

                            if service_name.replace("*", "") in service_list_name:
                                set_service_details(ssh_client, service_list_name, final_service_details)

                    else:
                        set_service_details(ssh_client, service_name, services_metrics)

            else:
                pass

        if len(services_metrics) > 0:
            metrics["Services"] = services_metrics

    except Exception:
        _logger.error()


def set_service_details(ssh_client, service, final_service_details):
    try:

        if service and len(service) > 0:

            command_to_execute = "service " + convert_to_str(service) + " status"

            partial_result = ssh_client.execute_command(command_to_execute)

            if partial_result == "" or partial_result == "\n" or partial_result == "\r\n":
                partial_result = ssh_client.execute_command("/sbin/service " + convert_to_str(service) + " status")

            row = {}

            row["Service"] = service

            pid = -1

            result = partial_result

            if partial_result and len(partial_result.split("\n")) > 1:

                for output in partial_result.split("\n"):

                    if service in output and "running" in output:
                        result = output

                        break

            if result and "pid" in result and "running." in result:

                for string_token in result.split("\n"):

                    if len(string_token) > 0:

                        if "running" in string_token:
                            row['Service Status'] = "Running"

                        if "pid" in string_token:

                            index = string_token.find("(pid") + 4

                            token = string_token.substring(index, string_token.find(")"))

                            tokens = token.strip().split(" ")

                            if len(tokens) > 0:
                                pid = convert_to_int(tokens[0])

                if pid == -1:
                    row['Service Status'] = "Stopped"

            elif result and  ("start/running" in result or "running" in result or "active (exited)" in result) and ("not running" not in result):

                row["Service Status"] = "Running"

            else:

                row["Service Status"] = "Stopped"

            if len(row) > 0:
                final_service_details.append(row)


    except Exception:
        _logger.error()


def get_services(ssh_client, services_metrics):
    service_list = set()

    try:

        output = ssh_client.execute_command("service --status-all")

        if output and len(convert_to_str(output)) > 0:

            services = output.split("\n")

            pattern = re.compile("(.*)\\s*is\\s*(\\w+)")

            for service_name in services:

                row = {}

                service_name = convert_to_str(service_name)

                if len(service_name) > 0 and "]" in service_name:

                    service_name = service_name[service_name.rfind("]") + 2:]

                    service_list.add(service_name)

                elif len(service_name) > 0 and (
                        "is disabled" in service_name or "is stopped" in service_name or "is running" in service_name):

                    matcher = pattern.match(service_name)

                    if matcher:

                        service_name = convert_to_str(matcher.group(1))

                        service_status = convert_to_str(matcher.group(2))

                        if "pid" in service_name and "running" in service_status:

                            service_name = re.sub("\(pid.*", "", service_name).replace("(", "")

                            row["Service"] = service_name

                            row['Service Status'] = "Running"

                        else:

                            row["Service"] = service_name

                            if "stopped" in service_status:
                                row["Service Status"] = "Stopped"


                            elif "abled." in service_status:

                                row["Service"] = service_name.replace("is d", "")

                                row["Service Status"] = "Disabled"

                            else:

                                row["Service Status"] = service_status

                        if len(row) == 2:
                            services_metrics.append(row)

            else:

                output = ssh_client.execute_command("systemctl list-unit-files | grep 'service' | awk '{print $1}'")

                if output and len(convert_to_str(output)) > 0 and (
                        convert_to_str(output) != "\n" or convert_to_str(output) != "\r\n"):

                    services = output.split("\n")

                    for service_name in services:
                        service_list.add(convert_to_str(service_name).replace(".service", ""))


    except Exception:
        _logger.error()

    return service_list


def set_password_expiry_days(ssh_client, metrics,user_name):
    try:

        output = ssh_client.execute_command("chage -l "+user_name)

        if output and len(output) > 0 and output != "\n":

            pattern = re.compile("Password expires\\s+:\\s+(.*)")

            for matcher in re.finditer(pattern, output):

                metrics["Password Expiry Date"] = convert_to_str(matcher.group(1))


    except Exception:

        _logger.error()


def convert_to_int(value):
    result = 0

    try:
        if value:

            result = int(round(float(str(value).replace("\r", "").strip())))

    except Exception:

        _logger.error()

    return result


def convert_to_float(value):
    result = 0

    try:
        if value:

            result = float(str(value).replace("\r", "").strip())

    except Exception:

        _logger.error()

    return result


def convert_to_str(value):
    result = 0

    try:
        if value:

            result = str(value).strip().replace("\r", "")

    except Exception:

        _logger.error()

    return result


def discover(ssh_client,context):

    try:

        if ssh_client:

            output =  ssh_client.execute_command("uname")

            if output and len(convert_to_str(output)) > 0:

                return output



    except Exception as exception:

        context['error-code'] = exception.__class__.__name__

        _logger.error()

    return None


if __name__ =="__main__":

    if len(sys.argv) > 0 and sys.argv[1] == 'discovery':

        context = {}

        try:
            context = json.loads(b64decode(sys.argv[2]))

            ssh_client = get_ssh_client(context, context['host'])

            if ssh_client and ssh_client.init():

                result = discover(ssh_client,context)

                if result is not None and len(result) > 0:

                    context['result'] = result
                else:

                    if ssh_client.exception is not None:

                        context['error-code'] = convert_to_str(ssh_client.exception)

                    else:
                        context['error-code'] = "Failed to discover "+context["host"]

                ssh_client.destroy()
            else:
                context['error-code'] = ssh_client.exception.__class__.__name__

            del context['password']

            del context['username']

        except Exception as exception:

            context['error-code'] = exception.__class__.__name__

            _logger.error()
        finally:
            print (json.dumps(context))
